/*************** Includes *****************************************************************************************/
#include "system.h"
#include "io_expander.h"

#ifdef LCD_IOEXPANDER1
#include "i2c1.h"
#endif

#ifdef LCD_IOEXPANDER2
#include "i2c2.h"
#endif

/*************** Data Declarations ********************************************************************************/
I2C_MESSAGE_STATUS I2C_Message_Status;      // variable to hold return STATUS from I2C Driver
uint8_t addr;                               // variable to hold IO Expander I2C address

extern bit TimerDone;                       // Flag for Timer to main program
extern bit Error;                           // Flag to indicate an error

/*******************************************************************************************************************
 * Function: void IO_ExpanderInitialize(uint8_t data, uint8_t address);
 * Overview: Initializes the I/O Expander
 * PreCondition: None
*******************************************************************************************************************/
void IO_ExpanderInitialize(uint8_t data, uint8_t address)
{
    addr = address;                 // save IO Expander I2C address
    
#ifdef LCD_IOEXPANDER1
    I2C1_Initialize();
#endif
    
#ifdef LCD_IOEXPANDER2
    I2C2_Initialize();
#endif
    
    IO_ExpanderWrite(data);
}

/*******************************************************************************************************************
 * Function: void IO_ExpanderWrite(uint8_t data);
 * Overview: Write Data to the I/O Expander using I2C
 * PreCondition: None
*******************************************************************************************************************/
void IO_ExpanderWrite(uint8_t data)
{
    uint8_t retry = 0; // variable for number of retries for device
    I2C_Message_Status = I2C_MESSAGE_PENDING;
    while (I2C_Message_Status != I2C_MESSAGE_FAIL)
    {
        I2C_MasterWrite(&data, 1, addr, &I2C_Message_Status);
        TMR1_StartTimer();                                      // start 1 sec timeout timer TMR1
        while (I2C_Message_Status == I2C_MESSAGE_PENDING)       // wait for message to be sent or status change
        {
            if (TimerDone == 1)
            {
                TimerDone = 0;                                  // set Flag to false
                I2C_Message_Status = I2C_MESSAGE_FAIL;          // set status to I2C_MESSAGE_FAIL
                break;   // break out of while loop, device is not responding
            }
        }
        if ((I2C_Message_Status == I2C_MESSAGE_COMPLETE) | (I2C_Message_Status == I2C_MESSAGE_FAIL))
            break;  // break out of while loop, message has been sent or failed
        // if status is I2Cx_MESSAGE_ADDRESS_NO_ACK or I2Cx_DATA_NO_ACK,
        // The device may be busy and needs more time for the last write so we can retry writing the data, 
        // this is why we use a while loop here
        if (retry == SLAVE_I2C_RETRY_MAX)   // check for max retry and skip this break
            break;
        else
            retry++;
    }
    if (!( I2C_Message_Status == I2C_MESSAGE_COMPLETE ))    Error = 1;
    TMR1_StopTimer();       // Stop 1 sec timeout timer TMR1, we don't need it any more
    TimerDone = 0;          // set Flag to false
}

/*******************************************************************************************************************
 * Function: uint8_t IO_ExpanderRead(uint8_t rs);
 * Overview: Read Data from the I/O Expander using I2C
 * PreCondition: None
*******************************************************************************************************************/
uint8_t IO_ExpanderRead(uint8_t rs)
{
    uint8_t data;           // data recieved from I/O Expander
    uint8_t lownib;         // data recieved from I/O Expander used for LOWER nibble
    uint8_t retry = 0;      // variable for number of retries for device
    
    // set Control line RS LOW or leave HIGH
    //IO_ExpanderWrite(Rw | rs);                  // RS LOW for Busy Flag and Address Counter
    //IO_ExpanderWrite(Rw | rs | En);             // Enable HIGH
    __delay_us(1);                              // Enable pulse must be >450ns
    //IO_ExpanderWrite(Rw | rs | (En & ~En));     // Enable LOW
    // now read IO Expander Port HIGH Nibble
    I2C_Message_Status = I2C_MESSAGE_PENDING;
    while (I2C_Message_Status != I2C_MESSAGE_FAIL)
    {
        I2C_MasterRead(&data, 1, addr, &I2C_Message_Status);
        TMR1_StartTimer();                                      // start 1 sec timeout timer TMR1
        while (I2C_Message_Status == I2C_MESSAGE_PENDING)       // wait for message to be sent or status change
        {
            if (TimerDone == true)
            {
                TimerDone = false;                              // set Flag to false
                I2C_Message_Status = I2C_MESSAGE_FAIL;          // set status to I2C_MESSAGE_FAIL
                break;   // break out of while loop, device is not responding
            }
        }
        if ((I2C_Message_Status == I2C_MESSAGE_COMPLETE) | (I2C_Message_Status == I2C_MESSAGE_FAIL))
            break;  // break out of while loop, message has not been rec'vd or failed
        // if status is I2Cx_MESSAGE_ADDRESS_NO_ACK or I2Cx_DATA_NO_ACK,
        // The device may be busy and needs more time for the last write so we can retry writing the data, 
        // this is why we use a while loop here
        if (retry == SLAVE_I2C_RETRY_MAX)   // check for max retry and skip this break
            break;
        else
            retry++;
    }
    if (!( I2C_Message_Status == I2C_MESSAGE_COMPLETE ))    Error = true;
    TMR1_StopTimer();       // Stop 1 sec timeout timer TMR1, we don't need it any more
    //IO_ExpanderWrite(Rw | rs | En);             // Enable HIGH
    __delay_us(1);                              // Enable pulse must be >450ns
    //IO_ExpanderWrite(Rw | rs | (En & ~En));     // Enable LOW
    // now read IO Expander Port LOW Nibble
    I2C_Message_Status = I2C_MESSAGE_PENDING;
    while (I2C_Message_Status != I2C_MESSAGE_FAIL)
    {
        I2C_MasterRead(&lownib, 1, addr, &I2C_Message_Status);
        TMR1_StartTimer();                                      // start 1 sec timeout timer TMR1
        while (I2C_Message_Status == I2C_MESSAGE_PENDING)       // wait for message to be sent or status change
        {
            if (TimerDone == true)
            {
                TimerDone = false;                      // set Flag to false
                I2C_Message_Status = I2C_MESSAGE_FAIL;          // set status to I2C_MESSAGE_FAIL
                break;   // break out of while loop, device is not responding
            }
        }
        if ((I2C_Message_Status == I2C_MESSAGE_COMPLETE) | (I2C_Message_Status == I2C_MESSAGE_FAIL))
            break;  // break out of while loop, message has not been rec'vd or failed
        // if status is I2Cx_MESSAGE_ADDRESS_NO_ACK or I2Cx_DATA_NO_ACK,
        // The device may be busy and needs more time for the last write so we can retry writing the data, 
        // this is why we use a while loop here
        if (retry == SLAVE_I2C_RETRY_MAX)   // check for max retry and skip this break
            break;
        else
            retry++;
    }
    if (!( I2C_Message_Status == I2C_MESSAGE_COMPLETE ))    Error = true;
    data = (data | (lownib >>4));
    TMR1_StopTimer();       // Stop 1 sec timeout timer TMR1, we don't need it any more
    return data;
}

void IO_ExpanderWaitBF(void)
{
    
}
/*************** End of File **************************************************************************************/