/*************** Includes *****************************************************************************************/
#include "system.h"
#include "lcd.h"

#ifndef LCD_IOPORT
#include "io_expander.h"
#endif

#ifdef LCD_IOPORT
#include "IO_Port.h"
#endif

#define IO_INIT 0x03                // Initialize LCD with RS and R/W lines HIGH
#define INSTWAIT 5                  // delay in mS for Instruction to execute
#define DATAWAIT 5                  // delay in mS for Data Write to execute

/*************** Defines ******************************************************************************************/
// commands
#define LCD_CLEARDISPLAY   0x01
#define LCD_RETURNHOME     0x02
#define LCD_ENTRYMODESET   0x04
#define LCD_DISPLAYCONTROL 0x08
#define LCD_CURDSPSHIFT    0x10
#define LCD_FUNCTIONSET    0x20
#define LCD_SETCGRAMADDR   0x40
#define LCD_SETDDRAMADDR   0x80
// flags for display entry mode
#define LCD_ENTRYDSPRIGHT  0x01
#define LCD_ENTRYDSPLEFT   0x03
#define LCD_ENTRYCURRIGHT  0x00
#define LCD_ENTRYCURLEFT   0x02
// flags for display on/off control
#define LCD_DISPLAYON      0x04
#define LCD_DISPLAYOFF     0x00
#define LCD_CURSORON       0x02
#define LCD_CURSOROFF      0x00
#define LCD_BLINKON        0x01
#define LCD_BLINKOFF       0x00
// flags for display/cursor shift
#define LCD_DISPLAYSHIFT   0x08
#define LCD_CURSORMOVE     0x00
#define LCD_MOVERIGHT      0x04
#define LCD_MOVELEFT       0x00
// flags for function set
#define LCD_8BITMODE       0x10
#define LCD_4BITMODE       0x00
#define LCD_2LINE          0x08
#define LCD_1LINE          0x00
#define LCD_5x10DOTS       0x04
#define LCD_5x8DOTS        0x00
// flags for backlight control
#define LCD_BACKLIGHT   0b00001011
#define LCD_NOBACKLIGHT 0b00000011
// RAM Addresses
#define LINE1_STADDR     0x00
#define LINE1_ENDDSPADDR 0x0F
#define LINE1_ENDRAMADDR 0x27
#define LINE2_STADDR     0x40
#define LINE2_ENDDSPADDR 0x4F
#define LINE2_ENDRAMADDR 0x67
// Display Control Flag Bits
#define BLINK       0x00
#define CURSOR      0x01
#define DISPLAY     0x02

/*******************************************************************************************************************
 * private data declarations
*******************************************************************************************************************/
/*******************************************************************************************************************
 * Summary: LCD Write Modes
 * Description: These define the modes for writting display
*******************************************************************************************************************/
typedef enum
{
    WRITEIO,                // Mode value for IO_Expansion or IO_Port command only
    WRITEINST,              // Mode value for Write Instruction
    WRITEDATA               // Mode value for Write Data
} WRITE_MODES;

/*******************************************************************************************************************
 * Summary: LCD Read Modes
 * Description: These define the modes for reading the display
*******************************************************************************************************************/
typedef enum
{
    READDATA,               // Mode value for Read Data
    READBFAC                // Mode value for Read BF and Address Counter
} READ_MODES;

uint8_t numChars;           // variable to hold the number of characters the display was init with .. cannot change
uint8_t numLines;           // variable to hold the number of lines of the display was init with .. cannot change
uint8_t AddrCounter;        // variable to hold currrent DDRAM or CGRAM Address Counter
uint8_t Control;            // variable to hold the state of the control pins RS, R/W, E and P3 (controls backlight)
uint8_t Display;            // variable to hold the state of the Display, Cursor & Blink ON or OFF states

/*******************************************************************************************************************
 * private macro functions
*******************************************************************************************************************/
#define setbit(var, bit) ((var) |= (1 << (bit)))
#define clrbit(var, bit) ((var) &= ~(1 << (bit)))
#define RS_SetHigh()  setbit(Control,0)
#define RS_SetLow()   clrbit(Control,0)
#define RW_SetHigh()  setbit(Control,1)
#define RW_SetLow()   clrbit(Control,1)
#define E_SetHigh()   setbit(Control,2)
#define E_SetLow()    clrbit(Control,2)
#define P3_SetHigh()  setbit(Control,3)
#define P3_SetLow()   clrbit(Control,3)
/*******************************************************************************************************************
 * private functions declarations
*******************************************************************************************************************/
void LCD_Write(uint8_t data, WRITE_MODES wmode);    // Write to the LCD Display in 4 bit mode
void LCD_Write8Bit(uint8_t data);                   // Write in 8 bit mode - used during Initializing
uint8_t LCD_Read(READ_MODES rmode);                 // Reads from the LCD Display in 4 bit mode
void LCD_TMR0CallBack(void);                        // process TMR0 call backs every 500 mS

/*******************************************************************************************************************
 * private functions definitions
*******************************************************************************************************************/
/*******************************************************************************************************************
 * Function: void LCD_Write(uint8_t data, WRITE_MODES wmode);
 * Overview: Write Data or Instruction to the LCD Display in 4 Bit Mode
 * PreCondition: None
*******************************************************************************************************************/
void LCD_Write(uint8_t data, WRITE_MODES wmode)
{
    uint8_t highnib, lownib;
    // break value into 2 nibbles for 4 bit transfer to LCD Display
    highnib = data;
    highnib = (highnib & 0xF0);
    lownib = data;
    lownib = (lownib<<4);
    
#ifndef LCD_IOPORT
    switch (wmode)
    {
        case WRITEIO:       // Write IO_Expansion or IO_Port command only
        {
            IO_ExpanderWrite((uint8_t)&data);
            break;
        }
        case WRITEINST:
        {
            RS_SetLow();
            RW_SetLow();
            IO_ExpanderWrite(Control);              // RS and R/W lines LOW
            E_SetHigh();
            IO_ExpanderWrite(Control);              // Enable HIGH
            IO_ExpanderWrite(highnib | Control);    // send HIGH Nibble with Enable HIGH
            __delay_us(2);                          // Enable pulse must be >450ns
            E_SetLow();
            IO_ExpanderWrite(highnib | Control);    // Enable LOW with HIGH Nibble
            __delay_us(2);                          // delay 2uS between HIGH nibble and LOW nibble
            // Send LOW Nibble now
            E_SetHigh();
            IO_ExpanderWrite(Control);              // Enable HIGH
            IO_ExpanderWrite(lownib | Control);     // send LOW Nibble with Enable HIGH
            __delay_us(2);                          // Enable pulse must be >450ns
            E_SetLow();
            IO_ExpanderWrite(highnib | Control);    // Enable LOW with LOW Nibble
            __delay_us(2);                          // delay 2uS between LOW nibble RS and R/W lines HIGH
            RS_SetHigh();
            RW_SetHigh();
            IO_ExpanderWrite(Control);              // RS and R/W lines HIGH
            __delay_ms(INSTWAIT);                   // delay for Instruction to execute
            break;
        }    
        case WRITEDATA:     // fix this and use Control variable
        {
            RW_SetLow();
            IO_ExpanderWrite(Control);                      // R/W line LOW
            // send high nibble
            E_SetHigh();
            IO_ExpanderWrite(Control);                      // Enable HIGH
            __delay_us(1);                                  // delay 1uS between E HIGH and Data
            IO_ExpanderWrite(highnib | Control);            // send HIGH Nibble with Enable HIGH
            __delay_us(2);                                  // Enable pulse must be >450ns
            E_SetLow();
            IO_ExpanderWrite(highnib | Control);            // Enable LOW with HIGH Nibble
            __delay_us(2);                                  // delay 2uS between HIGH nibble and LOW nibble
            // Send LOW Nibble now
            E_SetHigh();
            IO_ExpanderWrite(Control);                      // Enable HIGH
            __delay_us(1);                                  // delay 1uS between E HIGH and Data
            IO_ExpanderWrite(lownib | Control);             // send LOW Nibble with Enable HIGH
            __delay_us(2);                                  // Enable pulse must be >450ns
            E_SetLow();
            IO_ExpanderWrite(lownib | Control);             // Enable LOW with LOW Nibble
            __delay_us(2);                                  // delay 2uS between LOW nibble RS and R/W lines HIGH
            RW_SetHigh();
            IO_ExpanderWrite(Control);                      // R/W line HIGH
            __delay_ms(DATAWAIT);                           // delay for Data Write to execute
        }
    }
#endif
   
#ifdef LCD_IOPORT
    switch (wmode)
    {
        case WRITEIO:       // Write IO_Expansion or IO_Port command only
        {
            IO_PortWrite((uint8_t)data);
            break;
        }
        case WRITEINST:
        {
            RS_SetLow();
            RW_SetLow();
            IO_PortWrite(Control);                          // RS and R/W lines LOW
            // send high nibble
            E_SetHigh();
            IO_PortWrite(Control);                          // Enable HIGH
            __delay_us(1);                                  // delay 1uS between E HIGH and Data
            IO_PortWrite(highnib | Control);                // send HIGH Nibble with Enable HIGH
            __delay_us(2);                                  // Enable pulse must be >450ns
            E_SetLow();
            IO_PortWrite(highnib | Control);                // Enable LOW with HIGH Nibble
            __delay_us(2);                                  // delay 2uS between HIGH nibble and LOW nibble
            // Send LOW Nibble now
            E_SetHigh();
            IO_PortWrite(Control);                          // Enable HIGH
            __delay_us(1);                                  // delay 1uS between E HIGH and Data
            IO_PortWrite(lownib | Control);                 // send LOW Nibble with Enable HIGH
            __delay_us(2);                                  // Enable pulse must be >450ns
            E_SetLow();
            IO_PortWrite(lownib | Control);                 // Enable LOW with LOW Nibble
            __delay_us(2);                                  // delay 2uS between LOW nibble RS and R/W lines HIGH
            RS_SetHigh();
            RW_SetHigh();
            IO_PortWrite(Control);                          // RS and R/W lines HIGH
            __delay_ms(INSTWAIT);                           // delay for Instruction to execute
            break;
        }
        case WRITEDATA:
        {
            RW_SetLow();
            IO_PortWrite(Control);                          // R/W line LOW
            // send high nibble
            E_SetHigh();
            IO_PortWrite(Control);                          // Enable HIGH
            __delay_us(1);                                  // delay 1uS between E HIGH and Data
            IO_PortWrite(highnib | Control);                // send HIGH Nibble with Enable HIGH
            __delay_us(2);                                  // Enable pulse must be >450ns
            E_SetLow();
            IO_PortWrite(highnib | Control);                // Enable LOW with HIGH Nibble
            __delay_us(2);                                  // delay 2uS between HIGH nibble and LOW nibble
            // Send LOW Nibble now
            E_SetHigh();
            IO_PortWrite(Control);                          // Enable HIGH
            __delay_us(1);                                  // delay 1uS between E HIGH and Data
            IO_PortWrite(lownib | Control);                 // send LOW Nibble with Enable HIGH
            __delay_us(2);                                  // Enable pulse must be >450ns
            E_SetLow();
            IO_PortWrite(lownib | Control);                 // Enable LOW with LOW Nibble
            __delay_us(2);                                  // delay 2uS between LOW nibble RS and R/W lines HIGH
            RW_SetHigh();
            IO_PortWrite(Control);                          // R/W line HIGH
            __delay_ms(DATAWAIT);                           // delay for Data Write to execute
        }
    }
    
#endif
}

/*******************************************************************************************************************
 * Function: void LCD_Write8Bit(uint8_t data);
 * Overview: Write Instruction to the LCD Display in 8 Bit Mode
 * PreCondition: None
*******************************************************************************************************************/
void LCD_Write8Bit(uint8_t data)
{
#ifndef LCD_IOPORT
    RS_SetLow();
    RW_SetLow();
    IO_ExpanderWrite(Control);                      // RS and R/W lines LOW
    E_SetHigh();
    IO_ExpanderWrite(Control);                      // Enable HIGH
    __delay_us(1);                                  // delay 1uS between E HIGH and Data
    IO_ExpanderWrite(data | Control);               // send HIGH Nibble with Enable HIGH
    __delay_us(2);                                  // Enable pulse must be >450ns
    E_SetLow();
    IO_ExpanderWrite(data | Control);               // Enable LOW with HIGH Nibble
    __delay_us(2);                                  // delay for Data Hold
    RS_SetHigh();
    RW_SetHigh();
    IO_ExpanderWrite(Control);                      // RS and R/W lines HIGH
    __delay_ms(INSTWAIT);                           // delay for Instruction to execute
#endif

#ifdef LCD_IOPORT
    RS_SetLow();
    RW_SetLow();
    IO_PortWrite(Control);                          // RS and R/W lines LOW
    E_SetHigh();
    IO_PortWrite(Control);                          // Enable HIGH
    __delay_us(1);                                  // delay 1uS between E HIGH and Data
    IO_PortWrite(data | Control);                   // send HIGH Nibble with Enable HIGH
    __delay_us(2);                                  // Enable pulse must be >450ns
    E_SetLow();
    IO_PortWrite(data | Control);                   // Enable LOW with HIGH Nibble
    __delay_us(2);                                  // delay for Data Hold
    RS_SetHigh();
    RW_SetHigh();
    IO_PortWrite(Control);                          // RS and R/W lines HIGH
    __delay_ms(INSTWAIT);                           // delay for Instruction to execute
#endif
}

/*******************************************************************************************************************
 * Function: uint8_t LCD_Read(READ_MODES rmode);
 * Overview: Read Data from the LCD Display in 4 Bit Mode
 * PreCondition: None
*******************************************************************************************************************/
uint8_t LCD_Read(READ_MODES rmode)
{
    uint8_t data;
    READ_MODES mode;
    
#ifndef LCD_IOPORT
    switch (rmode)
    {
        case READDATA:
        {
            //data = IO_ExpanderRead(Rs);
            break;
        }
        case READBFAC:
        {
            //data = IO_ExpanderRead(Rs & ~Rs);
        }
    }
#endif

#ifdef LCD_IOPORT
    // add code for Port I/O Read
    IO_PortRead();
#endif
    return data;
}

/*******************************************************************************************************************
 * Function: void LCD_TMR0CallBack(void);
 * Overview: process TMR0 call backs every 500 mS
 * PreCondition: None
*******************************************************************************************************************/
 void LCD_TMR0CallBack(void)
 {
     scrollDisplayLeft();
 }

/*******************************************************************************************************************
 * User APIs
*******************************************************************************************************************/

// When the display powers up, it is configured as follows:
//
// 1. Display clear
// 2. Function set: 
//    DL = 1; 8-bit interface data 
//    N = 0; 1-line display 
//    F = 0; 5x8 dot character font 
// 3. Display on/off control: 
//    D = 0; Display off 
//    C = 0; Cursor off 
//    B = 0; Blinking off 
// 4. Entry mode set: 
//    I/D = 1; Increment by 1
//    S = 0; No shift 

/*******************************************************************************************************************
 * Function: void LCD_Initialize(uint8_t lcd_chars, uint8_t lcd_lines);
 * Overview: Initializes the LCD Display
 * PreCondition: None
*******************************************************************************************************************/
void LCD_Initialize(uint8_t lcd_chars, uint8_t lcd_lines)
{
    uint8_t data;               // data to send to interface device

    numChars = lcd_chars;       // save number of chars of the display
    numLines = lcd_lines;       // save number of lines of the display
    Control = IO_INIT;          // Initialize Control Lines RS and R/W HIGH .. E and P3 LOW
    TMR0_Initialize();          // Initialize TMR0 for scrolling messages
    TMR0_SetInterruptHandler(LCD_TMR0CallBack); // set callback function for scrolling messages
    
#ifndef LCD_IOPORT
    IO_ExpanderInitialize(Control, LCD_ADDR);   // Init I/O Expander Port with RS and R/W lines HIGH
#endif
    
#ifdef LCD_IOPORT
    IO_PortInitialize(Control);   // Init I/O Port with RS and R/W lines HIGH
#endif

/*******************************************************************************************************************
 * In order to insure the display gets set into the 4 bit mode, we need to send
 * a Function Set command three times to the display to insure it is in 8 bit mode, then send a Function Set
 * command to put it into 4 bit mode.
*******************************************************************************************************************/
    // according to datasheet, we need at least 40ms after power rises above 2.7V before sending commands.
    __delay_ms(50);
    data = (LCD_FUNCTIONSET | LCD_8BITMODE);
    LCD_Write8Bit(data);                            // send 1st Function Set command (8 bit mode)
    LCD_Write8Bit(data);                            // send 2ed Function Set command (8 bit mode)
    LCD_Write8Bit(data);                            // send 3rd Function Set command (8 bit mode)
    data = (LCD_FUNCTIONSET | LCD_4BITMODE);
    LCD_Write8Bit(data);                            // send Function Set command, set DL = 4 bit (8 bit mode)
    // now we are in 4 bit mode
    // now specify the number of display lines and character font .. can't change after this
    if (numLines == 2)
    {
        data = (LCD_FUNCTIONSET | LCD_2LINE | LCD_5x8DOTS);
    }
    else
    {
       data = (LCD_FUNCTIONSET | LCD_1LINE | LCD_5x8DOTS);
    }
    LCD_Write(data, WRITEINST);                     // send Function Set command, set number of lines and font
    data = (LCD_CLEARDISPLAY);
    LCD_Write(data, WRITEINST);                     // send Clear Display command
    data = (LCD_ENTRYMODESET | LCD_ENTRYCURLEFT);
    LCD_Write(data, WRITEINST);                     // send Entry Mode command
    setbit(Display, DISPLAY);
    data = (LCD_DISPLAYCONTROL | Display);
    LCD_Write(data, WRITEINST);                     // send Display Control command
    DisplayOn();                                    // Display ON
    BacklightOn();                                  // Backlight ON
}

/*******************************************************************************************************************
 * Function: void ClearDisplay(void);
 * Overview: Clears the Display
 * PreCondition: None
*******************************************************************************************************************/
void ClearDisplay(void)
{
    LCD_Write(LCD_CLEARDISPLAY, WRITEINST);
    AddrCounter = LINE1_STADDR;
}

/*******************************************************************************************************************
 * Function: void ReturnHome(void);
 * Overview: Set Curser Position to ZERO
 * PreCondition: None
*******************************************************************************************************************/
void ReturnHome(void)
{
    LCD_Write(LCD_RETURNHOME, WRITEINST);
    AddrCounter = LINE1_STADDR;
}

/*******************************************************************************************************************
 * Function: void SetCursorPos(uint8_t col, uint8_t row);
 * Overview: Set Curser Position to col, row
 * PreCondition: None
*******************************************************************************************************************/
void SetCursorPos(uint8_t col, uint8_t row)
{
    int row_offsets[] = { 0x00, 0x40 };
    // convert col, row to Address
    col = (col-1);                                                          // col is ZERO based
    row = (row-1);                                                          // row is ZERO based
    LCD_Write((LCD_SETDDRAMADDR | (col + row_offsets[row])), WRITEINST);    // set new Address
    AddrCounter = (LCD_SETDDRAMADDR | (col + row_offsets[row]));            // update local Address Counter
}

/*******************************************************************************************************************
 * Function: void DisplayOff(void);
 * Overview: Turn the display off (quickly)
 * PreCondition: None
*******************************************************************************************************************/
void DisplayOff(void)
{
    clrbit(Display, DISPLAY);
    LCD_Write((LCD_DISPLAYCONTROL | Display), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void DisplayOn(void);
 * Overview: Turn the display on (quickly)
 * PreCondition: None
*******************************************************************************************************************/
void DisplayOn(void)
{
    setbit(Display, DISPLAY);
    LCD_Write((LCD_DISPLAYCONTROL | Display), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void CursorOff(void);
 * Overview: Turns the underline cursor off
 * PreCondition: None
*******************************************************************************************************************/
void CursorOff(void)
{
    clrbit(Display, CURSOR);
    LCD_Write((LCD_DISPLAYCONTROL | Display), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void CursorOn(void);
 * Overview: Turns the underline cursor on
 * PreCondition: None
*******************************************************************************************************************/
void CursorOn(void)
{
    setbit(Display, CURSOR);
    LCD_Write((LCD_DISPLAYCONTROL | Display), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void BlinkOff(void);
 * Overview: Turn off the blinking cursor
 * PreCondition: None
*******************************************************************************************************************/
void BlinkOff(void)
{
    clrbit(Display, BLINK);
    LCD_Write((LCD_DISPLAYCONTROL | Display), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void BlinkOn(void);
 * Overview: Turn off the blinking cursor
 * PreCondition: None
*******************************************************************************************************************/
void BlinkOn(void)
{
    setbit(Display, BLINK);
    LCD_Write((LCD_DISPLAYCONTROL | Display), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void BacklightOff(void);
 * Overview: Turn off backlight
 * PreCondition: None
*******************************************************************************************************************/
void BacklightOff(void)
{
    P3_SetLow();
    LCD_Write(Control, WRITEIO);
}

/*******************************************************************************************************************
 * Function: void BacklightOn(void);
 * Overview: Turn on backlight
 * PreCondition: None
*******************************************************************************************************************/
void BacklightOn(void)
{
    P3_SetHigh();
    LCD_Write(Control, WRITEIO);
}

/*******************************************************************************************************************
 * Function: void scrollDisplayLeft(void);
 * Overview: Scrolls the display LEFT
 * PreCondition: None
*******************************************************************************************************************/
void scrollDisplayLeft(void)
{
    LCD_Write((LCD_CURDSPSHIFT | LCD_DISPLAYSHIFT | LCD_MOVELEFT), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void scrollDisplayRight(void);
 * Overview: Scrolls the display RIGHT
 * PreCondition: None
*******************************************************************************************************************/
void scrollDisplayRight(void)
{
    LCD_Write((LCD_CURDSPSHIFT | LCD_DISPLAYSHIFT | LCD_MOVERIGHT), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void CursorLeft(void);
 * Overview: Move the cursor LEFT one position without changing display data
 * PreCondition: None
*******************************************************************************************************************/
void CursorLeft(void)
{
    LCD_Write((LCD_CURDSPSHIFT | LCD_CURSORMOVE | LCD_MOVELEFT), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void CursorRight(void);
 * Overview: Move the cursor RIGHT one position without changing display data
 * PreCondition: None
*******************************************************************************************************************/
void CursorRight(void)
{
    LCD_Write((LCD_CURDSPSHIFT | LCD_CURSORMOVE | LCD_MOVERIGHT), WRITEINST);
}

/*******************************************************************************************************************
 * Function: void WriteData(uint8_t data);
 * Overview: Send data at the current DDRAM or CGRAM Address
 * Notes: User MUST set the address (either via SetCGAddr or SetDDAddr) last address set determines
 *        which memory data is written to. The Address Counter (AC) will INC or DEC according to mode set
 *        with the Entry mode set command previously issued. Set Address again to change memory (either CG or DD)
*******************************************************************************************************************/
void WriteData(uint8_t data)
{
    LCD_Write(data, WRITEDATA);             // Write data to LCD Display
    AddrCounter++;                          // increment Address Counter
    // check Address Counter for end of LCD Display line, set to next line or back to first as needed
    if (AddrCounter == (LINE1_ENDDSPADDR +1))
    {
        SetDDRAMAddr(LINE2_STADDR);
        AddrCounter = LINE2_STADDR;
    }
    if (AddrCounter == (LINE2_ENDDSPADDR +1))
    {
        SetDDRAMAddr(LINE1_STADDR);
        AddrCounter = LINE1_STADDR;
    }
}

/*******************************************************************************************************************
 * Function: void WriteDataMsg(uint8_t *data, uint8_t num);
 * Overview: Send block of data at the current DDRAM or CGRAM Address to the LCD Display
 * Notes: User MUST set the address (either via SetCGAddr or SetDDAddr) last address set determines
 *        which memory data is written to. The Address Counter (AC) will INC or DEC according to mode set
 *        with the Entry mode set command previously issued. Set Address again to change memory (either CG or DD)
 *        This is a BLOCKING function that returns when all data has been sent to the LCD Display
*******************************************************************************************************************/
void WriteDataMsg(uint8_t *data, uint8_t num)
{
    for (uint8_t cnt = 0; cnt < num; cnt++)
    {
        WriteData(data[cnt]);
    }
}

/*******************************************************************************************************************
 * Function: uint8_t ReadData(void);
 * Overview: Read Data Register at the current DDRAM or CGRAM Address from the LCD display
 * Notes: User MUST set the address (either via SetCGAddr or SetDDAddr) last address set determines
 *        which memory data is read from. The Address Counter (AC) will INC or DEC according to mode set
 *        with the Entry mode set command previously issued. Set Address again to change memory (either CG or DD)
*******************************************************************************************************************/
uint8_t ReadData(void)
{
    return LCD_Read(READDATA);
}

/*******************************************************************************************************************
 * Function: void SetCGRAMAddr(uint8_t addr);
 * Overview: set the Address Counter to addr
 * Notes: User MUST insure that the supplied addr is a valid CG RAM Address for this display
 *        This also set which memory (Display or CharGen) subsequent ReadData or WriteData functions use
*******************************************************************************************************************/
void SetCGRAMAddr(uint8_t addr)
{
    uint8_t data;                           // data to send to interface device
    
    data = (LCD_SETCGRAMADDR | addr);
    LCD_Write(data, WRITEINST);
}

/*******************************************************************************************************************
 * Function: void SetDDRAMAddr(uint8_t addr);
 * Overview: set the Address Counter to addr
 * Notes: User MUST insure that the supplied addr is a valid DD RAM Address for this display
 *        This also set which memory (Display or CharGen) subsequent ReadData or WriteData functions use
*******************************************************************************************************************/
void SetDDRAMAddr(uint8_t addr)
{
    uint8_t data;                           // data to send to interface device
    
    data = (LCD_SETDDRAMADDR | addr);
    LCD_Write(data, WRITEINST);
    AddrCounter = addr;
}

/*******************************************************************************************************************
 * Function: uint8_t GetDDRAMAddr(void);
 * Overview: get the current Address Counter
*******************************************************************************************************************/
uint8_t GetDDRAMAddr(void)
{
    return AddrCounter;
}

/*******************************************************************************************************************
 * Function: void WriteDDRAMData(uint8_t data);
 * Overview: Send data at the DDRAM for a Scrolling message
 * Notes: User MUST set the address with SetDDAddr and Entry mode to INC to set up this type of message
*******************************************************************************************************************/
void WriteDDRAMData(uint8_t data)
{
    LCD_Write(data, WRITEDATA);             // Write data to LCD Display
    AddrCounter++;                          // increment Address Counter
    // check Address Counter for end of LCD Display RAM, set to next line or back to first as needed
    if (AddrCounter == (LINE1_ENDRAMADDR +1))
    {
        SetDDRAMAddr(LINE2_STADDR);
        AddrCounter = LINE2_STADDR;
    }
    if (AddrCounter == (LINE2_ENDRAMADDR +1))
    {
        SetDDRAMAddr(LINE1_STADDR);
        AddrCounter = LINE1_STADDR;
    }
}

/*******************************************************************************************************************
 * Function: void WriteScrollMsg(uint8_t *data, uint8_t num);
 * Overview: Send block of data at the current DDRAM Address to the LCD Display
 * Notes: User MUST set the address with SetDDAddr and Entry mode to INC to set up this type of message
 *        This is a BLOCKING function that returns when all data has been sent to the LCD Display
*******************************************************************************************************************/
void WriteScrollMsg(uint8_t *data, uint8_t num)
{
    for (uint8_t cnt = 0; cnt < num; cnt++)
    {
        WriteDDRAMData(data[cnt]);
    }
}

/*******************************************************************************************************************
 * Function: void StartScrollMessage(void);
 * Overview: Starts a previously setup scroll message
 * Notes: User MUST setup message using WriteDDRAMData(data)
*******************************************************************************************************************/
void StartScrollMessage(void)
{
    INTCONbits.TMR0IE = 1;
}

/*******************************************************************************************************************
 * Function: void StopScrollMessage(void);
 * Overview: Stops a previously started scroll message
 * Notes: User MUST setup message using WriteDDRAMData(data) and start the scroll message with StartScrollMessage()
*******************************************************************************************************************/
void StopScrollMessage(void)
{
    INTCONbits.TMR0IE = 0;
}

/*******************************************************************************************************************
 * Function: uint8_t ReadBFAddr(void);
 * Overview: Read Address Counter (AC) and Busy Flag (BF)
*******************************************************************************************************************/
uint8_t ReadBFAddr(void)
{
    return LCD_Read(READBFAC);
}
/*************** End of File **************************************************************************************/