#ifndef LCD_H
#define LCD_H

/*******************************************************************************************************************
 * User APIs
*******************************************************************************************************************/
/*******************************************************************************************************************
 * Function: void LCD_Initialize(uint8_t lcd_chars, uint8_t lcd_lines);
 * Overview: Initializes the LCD Display
 * PreCondition: None
*******************************************************************************************************************/
void LCD_Initialize(uint8_t lcd_chars, uint8_t lcd_lines);

/*******************************************************************************************************************
 * Function: void ClearDisplay(void);
 * Overview: Clears the Display
 * PreCondition: None
*******************************************************************************************************************/
void ClearDisplay(void);

/*******************************************************************************************************************
 * Function: void ReturnHome(void);
 * Overview: Set Curser Position to ZERO
 * PreCondition: None
*******************************************************************************************************************/
void ReturnHome(void);

/*******************************************************************************************************************
 * Function: void SetCursorPos(uint8_t col, uint8_t row);
 * Overview: Set Curser Position to col, row
 * PreCondition: None
*******************************************************************************************************************/
void SetCursorPos(uint8_t col, uint8_t row);

/*******************************************************************************************************************
 * Function: void DisplayOff(void);
 * Overview: Turn the display off (quickly)
 * PreCondition: None
*******************************************************************************************************************/
void DisplayOff(void);

/*******************************************************************************************************************
 * Function: void DisplayOn(void);
 * Overview: Turn the display on (quickly)
 * PreCondition: None
*******************************************************************************************************************/
void DisplayOn(void);

/*******************************************************************************************************************
 * Function: void CursorOff(void);
 * Overview: Turns the underline cursor off
 * PreCondition: None
*******************************************************************************************************************/
void CursorOff(void);

/*******************************************************************************************************************
 * Function: void CursorOn(void);
 * Overview: Turns the underline cursor on
 * PreCondition: None
*******************************************************************************************************************/
void CursorOn(void);

/*******************************************************************************************************************
 * Function: void BlinkOff(void);
 * Overview: Turn off the blinking cursor
 * PreCondition: None
*******************************************************************************************************************/
void BlinkOff(void);

/*******************************************************************************************************************
 * Function: void BlinkOn(void);
 * Overview: Turn off the blinking cursor
 * PreCondition: None
*******************************************************************************************************************/
void BlinkOn(void);

/*******************************************************************************************************************
 * Function: void BacklightOff(void);
 * Overview: Turn off backlight
 * PreCondition: None
*******************************************************************************************************************/
void BacklightOff(void);

/*******************************************************************************************************************
 * Function: void BacklightOn(void);
 * Overview: Turn on backlight
 * PreCondition: None
*******************************************************************************************************************/
void BacklightOn(void);

/*******************************************************************************************************************
 * Function: void scrollDisplayLeft(void);
 * Overview: Scrolls the display LEFT
 * PreCondition: None
*******************************************************************************************************************/
void scrollDisplayLeft(void);

/*******************************************************************************************************************
 * Function: void scrollDisplayRight(void);
 * Overview: Scrolls the display RIGHT
 * PreCondition: None
*******************************************************************************************************************/
void scrollDisplayRight(void);

/*******************************************************************************************************************
 * Function: void CursorLeft(void);
 * Overview: Move the cursor LEFT one position without changing display data
 * PreCondition: None
*******************************************************************************************************************/
void CursorLeft(void);

/*******************************************************************************************************************
 * Function: void CursorRight(void);
 * Overview: Move the cursor RIGHT one position without changing display data
 * PreCondition: None
*******************************************************************************************************************/
void CursorRight(void);

/*******************************************************************************************************************
 * Function: void WriteData(uint8_t data);
 * Overview: Send data at the current DDRAM or CGRAM Address
 * Notes: User MUST set the address (either via SetCGAddr or SetDDAddr) last address set determines
 *        which memory data is written to. The Address Counter (AC) will INC or DEC according to mode set
 *        with the Entry mode set command previously issued. Set Address again to change memory (either CG or DD)
*******************************************************************************************************************/
void WriteData(uint8_t data);

/*******************************************************************************************************************
 * Function: void WriteDataMsg(uint8_t *data, uint8_t num);
 * Overview: Send block of data at the current DDRAM or CGRAM Address to the LCD Display
 * Notes: User MUST set the address (either via SetCGAddr or SetDDAddr) last address set determines
 *        which memory data is written to. The Address Counter (AC) will INC or DEC according to mode set
 *        with the Entry mode set command previously issued. Set Address again to change memory (either CG or DD)
 *        This is a BLOCKING function that returns when all data has been sent to the LCD Display
*******************************************************************************************************************/
void WriteDataMsg(uint8_t *data, uint8_t num);

/*******************************************************************************************************************
 * Function: uint8_t ReadData(void);
 * Overview: Read Data Register at the current DDRAM or CGRAM Address from the LCD display
 * Notes: User MUST set the address (either via SetCGAddr or SetDDAddr) last address set determines
 *        which memory data is read from. The Address Counter (AC) will INC or DEC according to mode set
 *        with the Entry mode set command previously issued. Set Address again to change memory (either CG or DD)
*******************************************************************************************************************/
uint8_t ReadData(void);

/*******************************************************************************************************************
 * Function: void SetCGRAMAddr(uint8_t addr);
 * Overview: set the Address Counter to addr
 * Notes: User MUST insure that the supplied addr is a valid CG RAM Address for this display
 *        This also set which memory (Display or CharGen) subsequent ReadData or WriteData functions use
*******************************************************************************************************************/
void SetCGRAMAddr(uint8_t addr);

/*******************************************************************************************************************
 * Function: void SetDDRAMAddr(uint8_t addr);
 * Overview: set the Address Counter to addr
 * Notes: User MUST insure that the supplied addr is a valid DD RAM Address for this display
 *        This also set which memory (Display or CharGen) subsequent ReadData or WriteData functions use
*******************************************************************************************************************/
void SetDDRAMAddr(uint8_t addr);

/*******************************************************************************************************************
 * Function: uint8_t GetDDRAMAddr(void);
 * Overview: get the current Address Counter
*******************************************************************************************************************/
uint8_t GetDDRAMAddr(void);

/*******************************************************************************************************************
 * Function: void WriteDDRAMData(uint8_t data);
 * Overview: Send data at the DDRAM for a Scrolling message
 * Notes: User MUST set the address with SetDDAddr and Entry mode to INC to set up this type of message
*******************************************************************************************************************/
void WriteDDRAMData(uint8_t data);

/*******************************************************************************************************************
 * Function: void WriteScrollMsg(uint8_t *data, uint8_t num);
 * Overview: Send block of data at the current DDRAM Address to the LCD Display
 * Notes: User MUST set the address with SetDDAddr and Entry mode to INC to set up this type of message
 *        This is a BLOCKING function that returns when all data has been sent to the LCD Display
*******************************************************************************************************************/
void WriteScrollMsg(uint8_t *data, uint8_t num);

/*******************************************************************************************************************
 * Function: void StartScrollMessage(void);
 * Overview: Starts a previously setup scroll message
 * Notes: User MUST setup message using WriteDDRAMData(data)
*******************************************************************************************************************/
void StartScrollMessage(void);

/*******************************************************************************************************************
 * Function: void StopScrollMessage(void);
 * Overview: Stops a previously started scroll message
 * Notes: User MUST setup message using WriteDDRAMData(data) and start the scroll message with StartScrollMessage()
*******************************************************************************************************************/
void StopScrollMessage(void);

/*******************************************************************************************************************
 * Function: uint8_t ReadBFAddr(void);
 * Overview: Read Address Counter (AC) and Busy Flag (BF)
*******************************************************************************************************************/
uint8_t ReadBFAddr(void);

#endif  /* LCD_H */
/*************** End of File **************************************************************************************/